unit Serie_datos;

{$mode objfpc}{$H+}

interface

uses
    xMatDefs, Classes, matreal, SysUtils, uAuxiliares, fddp, ufechas, Math,
    Uniform;


const
  VERSION_FORMATO_SERIES = 2;
type

  //TDAOf_ddp_VectDeMuestras = array of Tf_ddp_VectDeMuestras;
  //TMatOf_ddp_VectDeMuestras = array of TDAOf_ddp_VectDeMuestras;
  TDAOfVectR = array of TVectR;
  TIPO_SERIE = (SALIDA, ENTRADA);


    (* Clase para soporte de las series de datos *)
  TSeriesDeDatos = class
  public
    version: integer;
    archi: string;
    NSeries: integer;
    // Información de la primera muedtra (PM)
    PM_Anio, PM_Mes, PM_Dia, PM_Hora, PM_Minuto, PM_segundo: integer;
    FechaIni: TFecha;
    PeriodoDeMuestreo_horas: NReal;
    NPuntos, NPPorCiclo, Ncurvas: integer;
    nombresSeries: TStringList;
    series: TDAOfVectR;
    curvas_dem: TDAofVectR;
    tipo_serie: array of TIPO_SERIE;
{$IFDEF GRUPOS_POLARES}
    GruposPolares: TGruposPolares;
{$ENDIF}
    constructor CreateFromArchi(archi1: string);
    procedure WriteToArchi(archi1: string);
    procedure Free;
  end;
procedure Algoritmo_k_means(datos: TDAofVectR; nro_clusters, nro_iter:integer;tolerancia:Nreal; Fecha_Ini:TFecha);
implementation


constructor TSeriesDeDatos.CreateFromArchi(archi1: string);
var
  f: TextFile;
  r, serie: string;
  kserie, kpunto, kcurva, khora, k: integer;
  cnt_linea: integer;
  temp: NReal;

  procedure readln(var f: textfile; var s: string);
  begin
    system.readln(f, s);
    Inc(cnt_linea);
  end;

begin
  cnt_linea := 0;

  try

    inherited Create;
    self.archi := archi1;
    assignfile(f, archi1);
  {$I-}
    reset(f);
  {$I+}
    if ioresult <> 0 then
      raise Exception.Create('No puedo abrir el archivo: ' + archi1);
    uauxiliares.setSeparadoresGlobales;

    readln(f, r);
    if pos('VERSION_FORMATO_SERIES', r) <> 0 then
    begin
      nextpal(r);
      version := nextInt(r);
    end
    else
      version := 0;

    (* leemos los parámetros globales *)
    if (version > 0) then
    begin
      readln(f, r); // cantidad de series a tratar
      NSeries := nextInt(r);
    end
    else
      NSeries := nextInt(r);

    readln(f, r);

    PM_Anio := nextInt(r);
    PM_Mes := nextInt(r);
    PM_Dia := nextInt(r);
    PM_Hora := nextInt(r);
    PM_Minuto := nextInt(r);
    PM_Segundo := nextInt(r);
    readln(f, r);
    PeriodoDeMuestreo_horas := nextFloat(r);
    FechaIni:=TFecha.Create_AnioMesDia(PM_Anio, PM_Mes, PM_Dia);

    readln(f, r);
    NPuntos := nextInt(r);    // cantidad de puntos totales por serie
    readln(f, r);
    NPPorCiclo := nextInt(r); // cantidad de puntos en un ciclo



    setlength(self.tipo_serie, NSeries);
    if (version > 0) then
    begin
      readln(f, r);
      for kserie := 1 to NSeries do
      begin
        if r[kserie] = 'x' then
          tipo_serie[kserie - 1] := SALIDA
        else
          tipo_serie[kserie - 1] := ENTRADA;
      end;
    end
    else
    begin
      for kserie := 1 to NSeries - 1 do
        tipo_serie[kserie - 1] := SALIDA;
    end;

    readln(f, r); // encabezado de las series

    nombresSeries := TStringList.Create;
    for kserie := 0 to NSeries - 1 do
    begin
      serie := NextPal(r);
      nombresSeries.Add(serie);
    end;

    // Creamos e inicializamos el soporte para las series
    setlength(series, NSeries);
    for kserie := 1 to NSeries do
      series[kserie - 1] := TVectR.Create_Init(NPuntos);

    // Leemos los datos en las series $
    for kpunto := 1 to NPuntos do
    begin
      Read(f, k);
      for kserie := 1 to nSeries do
        Read(f, series[kserie - 1].pv[kpunto]);
      readln(f, r);
    end;

    // Creamos e inicializamos el soporte para las curvas de demanda.
    Ncurvas:= Round(NPuntos/24);
    setlength(curvas_dem, Ncurvas);
    for kcurva := 1 to Ncurvas do
      curvas_dem[kcurva - 1] := TVectR.Create_Init(24);

    // Creamos las curvas de demanda

    for kcurva:=1 to Ncurvas do
    begin
          begin
               for khora:=1 to 24 do
                   curvas_dem[kcurva-1].pv[khora]:= series[7].pv[(kcurva-1)*24+khora];
               // curvas_dem[kcurva-1].Writeconsola();
               curvas_dem[kcurva-1].HacerUnitario;
               //curvas_dem[kcurva-1].Writeconsola();
          end;
    end;



    uauxiliares.setSeparadoresLocales;
    CloseFile(f);
  except
    raise Exception.Create('Error leyendo línea: ' + IntToStr(cnt_linea));
  end;
end;

procedure TSeriesDeDatos.WriteToArchi(archi1: string);
var
  f: TextFile;
  r: string;
  kserie, kpunto: integer;
begin
  self.archi := archi;
  assignfile(f, archi);
  {$I-}
  rewrite(f);
  {$I+}
  if ioresult <> 0 then
    raise Exception.Create('No puedo abrir el archivo: ' + archi);
  uauxiliares.setSeparadoresGlobales;

  system.writeln(f, 'VERSION_FORMATO_SERIES:', #9, VERSION_FORMATO_SERIES);

  (* leemos los parámetros globales *)
  system.writeln(f, NSeries); // cantidad de series a tratar

  r := IntToStr(PM_Anio) + #9 + IntToStr(PM_Mes) + #9 + IntToStr(
    PM_Dia) + #9 + IntToStr(PM_Hora) + #9 + IntToStr(PM_Minuto) + #9 +
    IntToStr(PM_Segundo);
  system.writeln(f, r);

  writeln(f, PeriodoDeMuestreo_horas);


  system.writeln(f, NPuntos);    // cantidad de puntos totales por serie
  system.writeln(f, NPPorCiclo); // cantidad de puntos en un ciclo

  r := '';
  for kserie := 1 to NSeries do
  begin
    if tipo_serie[kserie - 1] = SALIDA then
      r := r + 'x'
    else
      r := r + 'y';
  end;
  system.writeln(f, r);

  r := '';
  for kserie := 0 to NSeries - 1 do
  begin
    if kserie > 0 then
      r := r + #9;
    r := r + nombresSeries[kserie];
  end;
  system.writeln(f, r); // encabezado de las series


  // escribimos los datos de las series
  for kpunto := 1 to NPuntos do
  begin
    system.Write(f, kpunto);
    for kserie := 1 to nSeries do
      system.Write(f, #9, series[kserie - 1].pv[kpunto]);
    system.writeln(f);
  end;
  uauxiliares.setSeparadoresLocales;
  CloseFile(f);

end;

procedure TSeriesDeDatos.Free;
var
  i: integer;
begin
  if nombresSeries <> nil then
    nombresSeries.Free;
  if series <> nil then
  begin
    for i := 0 to high(series) do
      series[i].Free;
    setlength(series, 0);
  end;

  if curvas_dem <> nil then
  begin
    for i := 0 to high(curvas_dem) do
      curvas_dem[i].Free;
    setlength(curvas_dem, 0);
  end;

  inherited Free;

end;

procedure Algoritmo_k_means(datos: TDAofVectR; nro_clusters,nro_iter:integer; tolerancia:Nreal; Fecha_Ini:TFecha);
var

  // Conjunto de vectores para almacenar los centroides y los vectores asociados
   Clusters, Centroides:TDAofVectR;
   kserie, Nserie, ktemp, k, j, i:integer;
   dist_clusters, Vect_temp, Vect_tol:TVectR;
   dist:NReal;
   kMin,kMax, centro: integer;
   Min,Max,tol:NReal;
   fsal: textfile;
   Fecha: TFecha;
begin
     randomize;
     randomize;

     Nserie:=length(datos);
     dist_clusters:=TVectR.Create_Init(nro_clusters);
     Vect_tol:=TVectR.Create_Init(nro_clusters);
     // cargo el centroide en cada uno de los cluster en forma aleatoria
     setlength(Clusters, nro_clusters);
     setlength(Centroides, nro_clusters);
     for kserie := 1 to nro_clusters do
     begin
         Clusters[kserie - 1] := TVectR.Create_Init(0);
         Centroides[kserie - 1] := TVectR.Create_Init(24);
         Centro:= round(Call_UNI*length(datos));
         Centroides[kserie-1]:=datos[Centro];
         writeln(Centro);
     end;

     // cargo los restantes datos en cada cluster dependiendo de la distancia a los centroides

     for kserie:=1 to nro_clusters do
       Clusters[kserie-1].Writeconsola();

     for k:= 1 to length(datos) do
     begin
         dist_clusters.Ceros;
         for kserie:=1 to nro_clusters do
             begin
               dist:= datos[k-1].distancia2(Centroides[kserie-1]);
               dist_clusters.pv[kserie]:= dist;
             end;
         dist_clusters.MinMax(kMin, kMax, Min, Max);
         Clusters[kmin-1].resize(Clusters[kmin-1].n+1);
         Clusters[kmin-1].pv[Clusters[kmin-1].n]:= k-1;
     end;



     Vect_temp:=TVectR.Create_Init(24);

     // Inicio las iteraciones

     i:=1;
     tol:=10;

     while (i<nro_iter) and (tol>tolerancia) do
       begin

         // calculo de los nuevos centroides
         for kserie:=1 to nro_clusters do
           begin
               Vect_temp.Ceros;
               Vect_tol.Ceros;
               for k:=1 to Clusters[kserie-1].n do
                 begin

                   ktemp:=round(Clusters[kserie-1].pv[k]);
                   Vect_temp.sum(datos[ktemp]);
                 end;
               Vect_temp.PorReal(1/Clusters[kserie-1].n);
               Vect_tol.pon_e(kserie-1,Vect_temp.distancia(Centroides[kserie - 1]));
               Centroides[kserie - 1].Copy(Vect_temp);
           end;
         Vect_tol.MinMax(kmin,kmax,min,max);
         tol:=max;


         // redistribuyo los puntos nuevamente
         for kserie := 1 to nro_clusters do
             Clusters[kserie - 1].resize(0);

         for k:= 1 to length(datos) do
         begin
             dist_clusters.Ceros;
             for kserie:=1 to nro_clusters do
                 begin
                   dist:= datos[k-1].distancia2(Centroides[kserie-1]);
                   dist_clusters.pv[kserie]:= dist;
                 end;
             dist_clusters.MinMax(kMin, kMax, Min, Max);
             Clusters[kmin-1].resize(Clusters[kmin-1].n+1);
             Clusters[kmin-1].pv[Clusters[kmin-1].n]:= k-1;
         end;
      writeln(i,#9,tol, '-----------------------------------------------');
      for kserie:=1 to nro_clusters do
       Clusters[kserie-1].Writeconsola();



       i:=i+1;
       end;



     assignfile( fsal, 'demandas.xlt' );
     rewrite( fsal );
     for kserie:=1 to nro_clusters do
      begin
           for j:=1 to Clusters[kserie-1].n do
            begin
               Fecha:=TFecha.Create_Clone(Fecha_Ini);
               Fecha.addDias(round(clusters[kserie-1].pv[j]));
               write(fsal, Fecha.AsStr);
               datos[round(clusters[kserie-1].pv[j])].WriteXLTsimple(fsal);
            end;
           //writeln(fsal);
           Centroides[kserie-1].WriteXLTSimple(fsal);
           writeln(fsal,'--------------------------');
      end;



     closefile( fsal );


     // descargo de memoria los vectores auxiliares
     dist_clusters.Free;
     Vect_temp.Free;
     Vect_tol.Free;

     if Clusters <> nil then
     begin
          for i := 0 to high(Clusters) do
              Clusters[i].Free;
          setlength(Clusters, 0);
     end;
     if Centroides <> nil then
     begin
          for i := 0 to high(Centroides) do
              Centroides[i].Free;
          setlength(Centroides, 0);
     end;


end;

end.

