{x$DEFINE OPTIMIZACION_ALFARO}
unit uRobotOptimizadorMultiCore;

interface

uses
{$IFDEF WIN32}
  ipcthrd,
{$ELSE}
  uEmuladorWinIPC, uKeyDir,
{$ENDIF}
  uSalasDeJuegoDistr, xMatDefs, classes, SysUtils;

var
  cntCalculando: Integer;
{$IFDEF WIN32}
  nomSemCntCalculando, nomEvPasoCompleto, nombreEvHiloProximoPaso: String;
{$ELSE}
  nomSemCntCalculando, nomEvPasoCompleto, nombreEvHiloProximoPaso: String;
{$ENDIF}

type
  TRobotOptimizadorMultiCore = class(TThread)
    public
      sala: TSalaDeJuego;
      estrellaIni, estrellaFin: Integer;
      darPasoAuto: boolean;
      iRobotEnLaLista: Integer;

      Constructor Create(sala: TSalaDeJuego; const costoFuturo: TMatOfNReal; darPasoAutomaticamente: boolean; iRobotEnLaLista: Integer);
      procedure Execute; override;
      procedure liberarSala;
  end;

//TODO definir un criterio para cuantos hilos usar
function getNHilos(nCores, nEstrellas, nSorteos, nMaxIteraciones: Integer) : Integer;

{$IFDEF LINUX}
procedure CrearArchiKey( s: string );
{$ENDIF}

implementation

function getNHilos(nCores, nEstrellas, nSorteos, nMaxIteraciones: Integer) : Integer;
{$IFDEF OPTIMIZACION_ALFARO}
var
  nPuntosACalcular: Integer;
{$ENDIF}
begin
{$IFNDEF OPTIMIZACION_ALFARO}
  result:= nCores;
{$ELSE}
  if nCores = 1 then
    result:= 1
  else
  begin
    nPuntosACalcular:= nEstrellas * nSorteos * (nMaxIteraciones + 1) div 2;
    if nCores = 2 then
    begin
      if nPuntosACalcular <= 100 then
        result:= 2
      else if nPuntosACalcular <= 1000 then
        result:= 4
      else if nPuntosACalcular <= 5000 then
        result:= 8
      else
        result:= 16
    end
    else
    begin
      if nPuntosACalcular <= 100 then
        result:= 4
      else if nPuntosACalcular <= 1000 then
        result:= 8
      else if nPuntosACalcular <= 5000 then
        result:= 16
      else
        result:= 32
    end
  end;
{$ENDIF}
end;

Constructor TRobotOptimizadorMultiCore.Create(sala: TSalaDeJuego; const costoFuturo: TMatOfNReal; darPasoAutomaticamente: boolean; iRobotEnLaLista: Integer);
begin
  inherited Create(true);
  //cargarSala se debe hacer secuencial
  self.sala:= TSalaDeJuego.cargarSala(sala.archiSala, true);
{$IFDEF ITERAR}
  self.sala.globs.NMAX_ITERACIONESDELPASO_OPT := sala.globs.NMAX_ITERACIONESDELPASO_OPT;
{$ENDIF}  
  self.sala.estabilizarInicio:= false;
  self.sala.inicializarOptimizacion(costoFuturo);
  self.iRobotEnLaLista:= iRobotEnLaLista;
  self.darPasoAuto:= darPasoAutomaticamente;
  self.FreeOnTerminate:= false;
  self.Resume;
end;

procedure TRobotOptimizadorMultiCore.Execute;
var
  semCntCalculando: TMutex;
{$IFDEF WIN32}
  evPasoCompleto, evProximoPaso: TEvent;
{$ELSE}
  evPasoCompleto: TEventTX;
  evProximoPaso: TEventRX;
{$ENDIF}
begin
{$IFDEF WIN32}
  semCntCalculando:= TMutex.Create(nomSemCntCalculando);
  evPasoCompleto:= TEvent.Create(nomEvPasoCompleto, false);
  evProximoPaso:= TEvent.Create(nombreEvHiloProximoPaso + IntToStr(iRobotEnLaLista), false);
{$ELSE}
  semCntCalculando:= TMutex.Create(nomSemCntCalculando, 1);
  evPasoCompleto:= TEventTX.Create(nomEvPasoCompleto, 1);
  evProximoPaso:= TEventRX.Create(nombreEvHiloProximoPaso, iRobotEnLaLista);
{$ENDIF}

  if not semCntCalculando.Get(6000) then
  begin
    writeln('TRobotOptimizadorMultiCore.Execute: No obtuve semCntCalculando 1');
    Terminate;
  end
  else
  begin
    cntCalculando:= cntCalculando - 1;
    if cntCalculando = 0 then
      evPasoCompleto.Signal;
    semCntCalculando.Release;
  end;

  while (sala.globs.kPaso_ > 0) and (not Terminated) do
  begin
    if not evProximoPaso.Wait(60000) then
    begin
      writeln('TRobotOptimizadorMultiCore.Execute: No me despertaron de evProximoPaso');
      Terminate;
      break;
    end;

    sala.calcularRangoEstrellas(estrellaIni, estrellaFin);
    if not semCntCalculando.Get(6000) then
    begin
      writeln('TRobotOptimizadorMultiCore.Execute: No obtuve semCntCalculando 2');
      Terminate;
      break;
    end;
    cntCalculando:= cntCalculando - 1;
    if cntCalculando = 0 then
      evPasoCompleto.Signal;
    semCntCalculando.Release;

    if darPasoAuto then
      sala.darPaso;
  end;
  semCntCalculando.Free;
  evPasoCompleto.Free;
  evProximoPaso.Free;
  //Esto esta mal aca. Debera llamarse en el destructor de la clase, pero
  //FreeOnTerminate = true no llama al metodo Free si esta redefinido y el
  //evento OnTerminate en Free Pascal se llama despus que el thread esta destruido
  //Si el Thread ejecuta bien no hay problema, pero sino hay que liberar la sala
  //a mano
  sala.Free;
  sala:= NIL;
end;

procedure TRobotOptimizadorMultiCore.liberarSala;
begin
  if sala <> NIL then
  begin
    sala.Free;
    sala:= NIL;
  end;
end;

{$IFDEF LINUX}
procedure CrearArchiKey( s: string );
var
 f: textfile;
begin
  //writeln('Creando archivo key ' + s);
  ForceDirectories(ExtractFilePath(s));
  AssignFile( f, s );
  Rewrite( f );
  writeln( f, 'ArchiKey: ',s,' - no borrar.' );
  closefile( f );
end;
{$ENDIF}

initialization
{$IFDEF WIN32}
  nomSemCntCalculando:= 'nomSemCntCalculando';
  nomEvPasoCompleto:= 'nomEvPasoCompleto';
  nombreEvHiloProximoPaso:= 'nombreEvHiloProximoPaso';
{$ELSE}
  nomSemCntCalculando:= uKeyDir.keyBaseDir + '/nomSemCntCalculando';
  nomEvPasoCompleto:= uKeyDir.keyBaseDir + '/nomEvPasoCompleto';
  nombreEvHiloProximoPaso:= uKeyDir.keyBaseDir + '/nombreEvHiloProximoPaso';
{$ENDIF}

end.
